/* BEGIN_ICS_COPYRIGHT4 ****************************************

Copyright (c) 2015, Intel Corporation

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

** END_ICS_COPYRIGHT4   ****************************************/
/* [ICS VERSION STRING: unknown] */


#ifndef __IBA_IB_IBNOTIFY_H__
#define __IBA_IB_IBNOTIFY_H__	(1)

#include "datatypes.h"
#include "stl_types.h"
#include "ib_debug.h"
#include "idebug_osd.h"
#include "statustext.h"
#include "ispinlock.h"
#include "ilist.h"
#include "imemory.h"
#include "imap.h"
#include "ibt.h"

#ifdef __cplusplus
extern "C" {
#endif


#define IBT_NOTIFY_TAG	MAKE_MEM_TAG(I,b,t,R)

//
// Forward declarations of Globals
//

typedef 
void
( *IB_ROOT_CALLBACK ) (
	IN	void			*Context
	);

//
// Group block
//
typedef struct _IBT_NOTIFY_GROUP {
	SPIN_LOCK			Lock;
	MAP_RES				Lists;	// key is Guid
	IB_ROOT_CALLBACK	RootCallback;
	boolean				Initialized;	// was it initialized
} IBT_NOTIFY_GROUP;

typedef struct _IBT_NOTIFY_USER_BLOCK {
	LIST_ITEM			ListItem;

	EUI64				Guid;			// Guid user wants events for, 0 for all
	IB_NOTIFY_CALLBACK	Callback;		// callback of user		
	void				*Context;		// user's context
	uint32				RefCount;		// Reference Count
	IB_NOTIFY_TYPE		EventMask;		// event mask		
	IB_NOTIFY_LOCKING	Locking;		// locking model

} IBT_NOTIFY_USER_BLOCK;


//
// Forward declarations of Functions
//

void
IbtNotifyGroup(
	IN	EUI64				Guid,			// Port/Node GUID
	IN	IB_NOTIFY_TYPE		EventType		// Type of event being reported
	);

void
IbtInitNotifyGroup(
	IN	IB_ROOT_CALLBACK	Callback
   );

void
IbtDestroyNotifyGroup(void);

boolean
IbtNotifyUserEvent(
	IN	void				*UserContext,	// User context passed by Register call
	IN	EUI64				Guid,			// Port/Node GUID
	IN	IB_NOTIFY_TYPE		EventType		// Type of event being reported
	);

//
// Debug specific info
//

#define	_DBG_PRINT_EPILOG(LEVEL,STRING)
#define	_DBG_PRINT_PROLOG(LEVEL,STRING)
#define	_DBG_ERROR_EPILOG(LEVEL,STRING)
#define	_DBG_ERROR_PROLOG(LEVEL,STRING)
#define	_DBG_FATAL_EPILOG(LEVEL,STRING)
#define	_DBG_FATAL_PROLOG(LEVEL,STRING)

#define	_DBG_LVL_IBT_API		0x00002000

#ifdef __cplusplus
};
#endif

#endif	// __IBA_IB_IBNOTIFY_H__
