#!/bin/bash
# BEGIN_ICS_COPYRIGHT8
# 
# Copyright (c) 2015, Intel Corporation
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 
# END_ICS_COPYRIGHT8

FM_CMD="/usr/lib/opa-fm/bin/fm_cmd"

cmd=(
	"smPerfDebug 100"
	"saPerfDebug 100"
	"saRmppDebug 100"
	"smForceSweep"
	"smRestorePriority"
	"smShowCounters"
	"smResetCounters"
	"smStateDump"
	"smLogLevel 4"
	"smLogLevel 2"
	"smLogMode 1"
	"smLogMode 0"
	"smLogMask 0x7ff"
	"smLogMask 0x1ff"
	"feLogLevel 4"
	"feLogLevel 2"
	"feLogMode 1"
	"feLogMode 0"
	"pmShowCounters"
	"pmResetCounters"
	"pmDebug 1"
	"pmDebug 0"
	"pmRmppDebug 1"
	"pmRmppDebug 0"
	"feDebug 1"
	"feDebug 0"
	"feRmppDebug 1"
	"feRmppDebug 0"
	"smForceRebalance"
	"smAdaptiveRouting"
	"smForceAttributeRewrite"
	"smSkipAttrWrite"
	"smPauseSweeps"
	"smResumeSweeps"
	"smPerfDebug 0"
	"saPerfDebug 0"
	"saRmppDebug 0"
	"smLooptestStart 5"
	"smLooptestStop"
	"smLooptestFastModeStart"
	"smLooptestStop"
)

echo "testing ${#cmd[@]} commands."

for i in `seq 0 $((${#cmd[@]} - 1))`
do
	echo "Testing ${cmd[$i]}...."
	$FM_CMD ${cmd[$i]}
	echo "-----------------------------------------"
done
